@@ -1,6 +1,7 @@
 package operator
 
 import (
+	"context"
 	"errors"
 	"fmt"
 	"sort"
@@ -16,12 +17,12 @@ import (
 	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/client-go/tools/cache"
 
-	metering "github.com/operator-framework/operator-metering/pkg/apis/metering/v1"
-	meteringUtil "github.com/operator-framework/operator-metering/pkg/apis/metering/v1/util"
-	"github.com/operator-framework/operator-metering/pkg/hive"
-	"github.com/operator-framework/operator-metering/pkg/operator/reporting"
-	"github.com/operator-framework/operator-metering/pkg/operator/reportingutil"
-	"github.com/operator-framework/operator-metering/pkg/util/slice"
+	metering "github.com/kube-reporting/metering-operator/pkg/apis/metering/v1"
+	meteringUtil "github.com/kube-reporting/metering-operator/pkg/apis/metering/v1/util"
+	"github.com/kube-reporting/metering-operator/pkg/hive"
+	"github.com/kube-reporting/metering-operator/pkg/operator/reporting"
+	"github.com/kube-reporting/metering-operator/pkg/operator/reportingutil"
+	"github.com/kube-reporting/metering-operator/pkg/util/slice"
 )
 
 const (
@@ -66,7 +67,7 @@ func init() {
 	prometheus.MustRegister(generateReportDurationHistogram)
 }
 
-func (op *Reporting) runReportWorker() {
+func (op *defaultReportingOperator) runReportWorker() {
 	logger := op.logger.WithField("component", "reportWorker")
 	logger.Infof("Report worker started")
 	const maxRequeues = 5
@@ -74,7 +75,7 @@ func (op *Reporting) runReportWorker() {
 	}
 }
 
-func (op *Reporting) syncReport(logger log.FieldLogger, key string) error {
+func (op *defaultReportingOperator) syncReport(logger log.FieldLogger, key string) error {
 	namespace, name, err := cache.SplitMetaNamespaceKey(key)
 	if err != nil {
 		logger.WithError(err).Errorf("invalid resource key :%s", key)
@@ -90,8 +91,22 @@ func (op *Reporting) syncReport(logger log.FieldLogger, key string) error {
 		}
 		return err
 	}
+
+	if report.Spec.Expiration != nil {
+		logger = logger.WithFields(log.Fields{"expiration": report.Spec.Expiration.Duration})
+	}
+
+	// TODO: we may want to move some of the validation checks that are
+	// in op.handleExpiredReport and op.handleReport such that we're not
+	// performing an expensive operation like creating a deep copy if we
+	// know that we're not going to be potentially modifying anything.
 	sr := report.DeepCopy()
 
+	err = op.handleExpiredReport(logger, sr, time.Now())
+	if err != nil {
+		return err
+	}
+
 	if report.DeletionTimestamp != nil {
 		_, err = op.removeReportFinalizer(sr)
 		return err
@@ -192,7 +207,7 @@ func getSchedule(reportSched *metering.ReportSchedule) (reportSchedule, error) {
 	return cron.Parse(cronSpec)
 }
 
-func (op *Reporting) handleReport(logger log.FieldLogger, report *metering.Report) error {
+func (op *defaultReportingOperator) handleReport(logger log.FieldLogger, report *metering.Report) error {
 	if op.cfg.EnableFinalizers && reportNeedsFinalizer(report) {
 		var err error
 		report, err = op.addReportFinalizer(report)
@@ -339,17 +354,65 @@ func getReportPeriod(now time.Time, logger log.FieldLogger, report *metering.Rep
 	return reportPeriod, nil
 }
 
+func (op *defaultReportingOperator) handleExpiredReport(logger log.FieldLogger, report *metering.Report, now time.Time) error {
+	// check if the Report is being deleted already
+	if report.DeletionTimestamp != nil {
+		logger.Warnf("report was already marked for deletion")
+		return nil
+	}
+	// check if the Report is past its retention time
+	if reportExpired := isReportExpired(op.logger, report, now); !reportExpired {
+		logger.Debugf("the %s Report in the %s namespace has not yet reached the expiration date", report.Name, report.Namespace)
+		return nil
+	}
+	// check if the Report is used by any other Report or ReportQuery, if not delete it
+	if reportIsNotInput := isReportNotUsedAsInput(logger, report, op); reportIsNotInput {
+		newDeletionTimestamp := metav1.Now()
+		report.SetDeletionTimestamp(&newDeletionTimestamp)
+		err := op.meteringClient.MeteringV1().Reports(report.Namespace).
+			Delete(context.TODO(), report.Name, metav1.DeleteOptions{})
+		if apierrors.IsNotFound(err) {
+			logger.Infof("report: %s, not deleted because it was not found at time delete attempted", report.Name)
+			return nil
+		}
+		if err != nil {
+			return fmt.Errorf("failed to delete the expired report: %s. err: %s", report.Name, err)
+		}
+		logger.Infof("deleted Report: %s in the Namespace: %s because it reached the expiration time",
+			report.Name, report.Namespace)
+		op.eventRecorder.Event(report, v1.EventTypeNormal, "ExpiredReportHasBeenDeleted",
+			fmt.Sprintf("Deleted the %s Report as the configured expiration date has passed", report.Name))
+
+		return nil
+	}
+	// if here, warn about dependency before return
+	logger.Warnf("report: %s, would be deleted because expired, but is depended on", report.Name)
+	op.eventRecorder.Event(report, v1.EventTypeWarning, "ExpiredReportHasDependencies",
+		fmt.Sprintf("Skipping the deletion of the %s Report as other resources are dependent on it, "+
+			"despite reaching the desired expiration date.", report.Name))
+
+	return nil
+}
+
 // runReport takes a report, and generates reporting data
 // according the report's schedule. If the next scheduled reporting period
 // hasn't elapsed, runReport will requeue the resource for a time when
 // the period has elapsed.
-func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report) error {
+func (op *defaultReportingOperator) runReport(logger log.FieldLogger, report *metering.Report) error {
+	// check if the report we're currently processing is considered
+	// "expired". If true, exit early and requeue that object so
+	// op.syncReport calls the proper handler for this resource.
+	if reportExpired := isReportExpired(logger, report, time.Now()); reportExpired {
+		logger.Infof("requeueing report that has reached its expiration date during the op.runReport method")
+		op.enqueueReport(report)
+		return nil
+	}
+
 	// check if the report was previously finished; store result in bool
 	if reportFinished := isReportFinished(logger, report); reportFinished {
 		return nil
 	}
 
-	runningCond := meteringUtil.GetReportCondition(report.Status, metering.ReportRunning)
 	queryGetter := reporting.NewReportQueryListerGetter(op.reportQueryLister)
 
 	// validate that Report contains valid Spec fields
@@ -458,7 +521,7 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 				},
 			},
 		}
-		_, err = op.meteringClient.MeteringV1().ReportDataSources(report.Namespace).Create(newReportDataSource)
+		_, err = op.meteringClient.MeteringV1().ReportDataSources(report.Namespace).Create(context.TODO(), newReportDataSource, metav1.CreateOptions{})
 		if err != nil {
 			if apierrors.IsAlreadyExists(err) {
 				logger.Infof("ReportDataSource %s already exists", dataSourceName)
@@ -469,7 +532,7 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 		logger.Infof("created PrestoTable ReportDataSource %s", dataSourceName)
 
 		report.Status.TableRef = v1.LocalObjectReference{Name: hiveTable.Name}
-		report, err = op.meteringClient.MeteringV1().Reports(report.Namespace).Update(report)
+		report, err = op.meteringClient.MeteringV1().Reports(report.Namespace).Update(context.TODO(), report, metav1.UpdateOptions{})
 		if err != nil {
 			logger.WithError(err).Errorf("unable to update Report status with tableName")
 			return err
@@ -484,10 +547,15 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 		}
 	}
 
-	var runningMsg, runningReason string
+	runningCond := meteringUtil.GetReportCondition(report.Status, metering.ReportRunning)
+
+	var (
+		runningMsg    string
+		runningReason string
+	)
 	if report.Spec.RunImmediately {
 		runningReason = meteringUtil.RunImmediatelyReason
-		runningMsg = fmt.Sprintf("Report %s scheduled: runImmediately=true bypassing reporting period [%s to %s].", report.Name, reportPeriod.periodStart, reportPeriod.periodEnd)
+		runningMsg = fmt.Sprintf("Report %s scheduled: runImmediately=true and reporting period [%s to %s].", report.Name, reportPeriod.periodStart, reportPeriod.periodEnd)
 	} else {
 		// Check if it's time to generate the report
 		if reportPeriod.periodEnd.After(now) {
@@ -690,7 +758,11 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 	if report.Spec.ReportingEnd != nil && report.Status.LastReportTime.Time.Equal(report.Spec.ReportingEnd.Time) {
 		msg := fmt.Sprintf("Report has finished reporting. Report has reached the configured spec.reportingEnd: %s", report.Spec.ReportingEnd.Time)
 		runningCond := meteringUtil.NewReportCondition(metering.ReportRunning, v1.ConditionFalse, meteringUtil.ReportFinishedReason, msg)
-		meteringUtil.SetReportCondition(&report.Status, *runningCond)
+
+		if condErr := meteringUtil.SetReportCondition(&report.Status, *runningCond); condErr != nil {
+			return condErr
+		}
+
 		logger.Infof(msg)
 	} else if report.Spec.Schedule != nil {
 		// determine the next reportTime, if it's not a run-once report and then
@@ -712,13 +784,17 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 
 		waitMsg := fmt.Sprintf("Next scheduled report period is [%s to %s]. next run time is %s.", reportPeriod.periodStart, reportPeriod.periodEnd, nextRunTime)
 		runningCond := meteringUtil.NewReportCondition(metering.ReportRunning, v1.ConditionFalse, meteringUtil.ReportingPeriodWaitingReason, waitMsg)
-		meteringUtil.SetReportCondition(&report.Status, *runningCond)
+
+		if condErr := meteringUtil.SetReportCondition(&report.Status, *runningCond); condErr != nil {
+			return condErr
+		}
+
 		logger.Infof(waitMsg+". waiting %s", waitTime)
 		op.enqueueReportAfter(report, waitTime)
 	}
 
 	// Update the status
-	report, err = op.meteringClient.MeteringV1().Reports(report.Namespace).Update(report)
+	report, err = op.meteringClient.MeteringV1().Reports(report.Namespace).Update(context.TODO(), report, metav1.UpdateOptions{})
 	if err != nil {
 		logger.WithError(err).Errorf("unable to update Report status")
 		return err
@@ -733,6 +809,60 @@ func (op *Reporting) runReport(logger log.FieldLogger, report *metering.Report)
 	return nil
 }
 
+// We check first for Reports depending on this Report, return if any, and then check ReportQueries depending on Report
+func isReportNotUsedAsInput(logger log.FieldLogger, report *metering.Report, op *defaultReportingOperator) bool {
+	// Consider Reports referencing this report in the namespace of this report
+	reports, err := op.reportLister.Reports(report.Namespace).List(labels.Everything())
+	if err != nil {
+		logger.Errorf("unable to determine list of reports that might be input for report: %s", report.Name)
+		return false
+	}
+	for _, report := range reports {
+		// If Report has no spec.inputs it can't be depending on this report.
+		if report.Spec.Inputs == nil {
+			continue
+		}
+		if does, depReport := op.reportsDependOnThisReport(logger, report, reports); does {
+			logger.Infof(
+				"report %s exists that uses report %s as input, "+
+					"will not delete though retention period has expired", depReport, report.Name)
+			return false
+		}
+	}
+	// Consider ReportQueries referencing this report in the namespace of this report
+	reportQueries, err := op.reportQueryLister.ReportQueries(report.Namespace).List(labels.Everything())
+	if err != nil {
+		logger.Errorf("unable to determine list of ReportQueries that might be input for report: %s", report.Name)
+		return false
+	}
+	for _, reportQuery := range reportQueries {
+		// If ReportQuery has no spec.inputs it can't be depending on this report.
+		if reportQuery.Spec.Inputs == nil {
+			continue
+		}
+
+		if does, depQuery := op.reportQueriesDependOnThisReport(logger, report, reportQueries); does {
+			logger.Infof("ReportQuery, %s exists that uses report %s as input, "+
+				"will not delete though retention period has expired", depQuery, report.Name)
+			return false
+		}
+	}
+	return true
+}
+
+func isReportExpired(logger log.FieldLogger, report *metering.Report, currentTime time.Time) bool {
+	if report.Spec.Expiration != nil {
+		currentTimeNano := currentTime.UTC().Truncate(time.Millisecond).UnixNano()
+		reportCreationNano := report.ObjectMeta.CreationTimestamp.UTC().Truncate(time.Millisecond).UnixNano()
+		retentionAmountNano := report.Spec.Expiration.Nanoseconds()
+		if currentTimeNano > (reportCreationNano + retentionAmountNano) {
+			return true
+		}
+		logger.Infof("curr: %d, creat: %d, retain: %d", currentTimeNano, reportCreationNano, retentionAmountNano)
+	}
+	return false
+}
+
 func getRunOnceReportPeriod(report *metering.Report) (*reportPeriod, error) {
 	if report.Spec.ReportingEnd == nil || report.Spec.ReportingStart == nil {
 		return nil, fmt.Errorf("run-once reports must have both ReportingEnd and ReportingStart")
@@ -773,9 +903,9 @@ func convertDayOfWeek(dow string) (int, error) {
 	return 0, fmt.Errorf("invalid day of week: %s", dow)
 }
 
-func (op *Reporting) addReportFinalizer(report *metering.Report) (*metering.Report, error) {
+func (op *defaultReportingOperator) addReportFinalizer(report *metering.Report) (*metering.Report, error) {
 	report.Finalizers = append(report.Finalizers, reportFinalizer)
-	newReport, err := op.meteringClient.MeteringV1().Reports(report.Namespace).Update(report)
+	newReport, err := op.meteringClient.MeteringV1().Reports(report.Namespace).Update(context.TODO(), report, metav1.UpdateOptions{})
 	logger := op.logger.WithFields(log.Fields{"report": report.Name, "namespace": report.Namespace})
 	if err != nil {
 		logger.WithError(err).Errorf("error adding %s finalizer to Report: %s/%s", reportFinalizer, report.Namespace, report.Name)
@@ -785,12 +915,12 @@ func (op *Reporting) addReportFinalizer(report *metering.Report) (*metering.Repo
 	return newReport, nil
 }
 
-func (op *Reporting) removeReportFinalizer(report *metering.Report) (*metering.Report, error) {
+func (op *defaultReportingOperator) removeReportFinalizer(report *metering.Report) (*metering.Report, error) {
 	if !slice.ContainsString(report.ObjectMeta.Finalizers, reportFinalizer, nil) {
 		return report, nil
 	}
 	report.Finalizers = slice.RemoveString(report.Finalizers, reportFinalizer, nil)
-	newReport, err := op.meteringClient.MeteringV1().Reports(report.Namespace).Update(report)
+	newReport, err := op.meteringClient.MeteringV1().Reports(report.Namespace).Update(context.TODO(), report, metav1.UpdateOptions{})
 	logger := op.logger.WithFields(log.Fields{"report": report.Name, "namespace": report.Namespace})
 	if err != nil {
 		logger.WithError(err).Errorf("error removing %s finalizer from Report: %s/%s", reportFinalizer, report.Namespace, report.Name)
@@ -804,12 +934,14 @@ func reportNeedsFinalizer(report *metering.Report) bool {
 	return report.ObjectMeta.DeletionTimestamp == nil && !slice.ContainsString(report.ObjectMeta.Finalizers, reportFinalizer, nil)
 }
 
-func (op *Reporting) updateReportStatus(report *metering.Report, cond *metering.ReportCondition) (*metering.Report, error) {
-	meteringUtil.SetReportCondition(&report.Status, *cond)
-	return op.meteringClient.MeteringV1().Reports(report.Namespace).Update(report)
+func (op *defaultReportingOperator) updateReportStatus(report *metering.Report, cond *metering.ReportCondition) (*metering.Report, error) {
+	if condErr := meteringUtil.SetReportCondition(&report.Status, *cond); condErr != nil {
+		return report, condErr
+	}
+	return op.meteringClient.MeteringV1().Reports(report.Namespace).Update(context.TODO(), report, metav1.UpdateOptions{})
 }
 
-func (op *Reporting) setReportStatusInvalidReport(report *metering.Report, msg string) error {
+func (op *defaultReportingOperator) setReportStatusInvalidReport(report *metering.Report, msg string) error {
 	logger := op.logger.WithFields(log.Fields{"report": report.Name, "namespace": report.Namespace})
 	// don't update unless the validation error changes
 	if runningCond := meteringUtil.GetReportCondition(report.Status, metering.ReportRunning); runningCond != nil && runningCond.Status == v1.ConditionFalse && runningCond.Reason == meteringUtil.InvalidReportReason && runningCond.Message == msg {
@@ -828,11 +960,11 @@ func GetReportQueryForReport(report *metering.Report, queryGetter reporting.Repo
 	return queryGetter.GetReportQuery(report.Namespace, report.Spec.QueryName)
 }
 
-func (op *Reporting) getReportDependencies(report *metering.Report) (*reporting.ReportQueryDependencies, error) {
+func (op *defaultReportingOperator) getReportDependencies(report *metering.Report) (*reporting.ReportQueryDependencies, error) {
 	return op.getQueryDependencies(report.Namespace, report.Spec.QueryName, report.Spec.Inputs)
 }
 
-func (op *Reporting) queueDependentReportsForReport(report *metering.Report) error {
+func (op *defaultReportingOperator) queueDependentReportsForReport(report *metering.Report) error {
 	// Look for all reports in the namespace
 	reports, err := op.reportLister.Reports(report.Namespace).List(labels.Everything())
 	if err != nil {
@@ -858,12 +990,51 @@ func (op *Reporting) queueDependentReportsForReport(report *metering.Report) err
 	return nil
 }
 
+// for each report in the namespace, find ones that depend on the report passed into the function
+func (op *defaultReportingOperator) reportsDependOnThisReport(logger log.FieldLogger, report *metering.Report, reports []*metering.Report) (bool, string) {
+	var depReportName string
+	for _, otherReport := range reports {
+		deps, err := op.getReportDependencies(otherReport)
+		if err != nil {
+			logger.Errorf("was unable to get dependencies for report: %s, while looking for dependent reports", report.Name)
+			return true, depReportName
+		}
+		// If a Report has a dependency on the passed in report, we're done
+		for _, dep := range deps.Reports {
+			if dep.Name == report.Name {
+				return true, dep.Name
+			}
+		}
+	}
+	return false, depReportName
+}
+
+// for each ReportQuery in the namespace, find ones that depend on the report passed into the function
+func (op *defaultReportingOperator) reportQueriesDependOnThisReport(logger log.FieldLogger, report *metering.Report, reportQueries []*metering.ReportQuery) (bool, string) {
+	var depQueryName string
+	// for each report in the namespace, find queries that depend on the report passed into the function.
+	for _, reportQuery := range reportQueries {
+		deps, err := op.getQueryDependencies(reportQuery.Namespace, reportQuery.Name, nil)
+		if err != nil {
+			logger.Errorf("was unable to get dependencies for report: %s, while looking for dependent queries")
+			return true, ""
+		}
+		// If this ReportQuery has a dependency on the passed in report, we're done
+		for _, dep := range deps.Reports {
+			if dep.Name == report.Name {
+				return true, dep.Name
+			}
+		}
+	}
+	return false, depQueryName
+}
+
 // queueDependentReportQueriesForReport will queue all
 // ReportQueries in the namespace which have a dependency on the
 // report
-func (op *Reporting) queueDependentReportQueriesForReport(report *metering.Report) error {
+func (op *defaultReportingOperator) queueDependentReportQueriesForReport(report *metering.Report) error {
 	queryLister := op.meteringClient.MeteringV1().ReportQueries(report.Namespace)
-	queries, err := queryLister.List(metav1.ListOptions{})
+	queries, err := queryLister.List(context.TODO(), metav1.ListOptions{})
 	if err != nil {
 		return err
 	}
